#define WINVER 0x0500 // for IDC_HAND
#include "statichyperlink.h"

#include <windowsx.h>

#include "resource.h"

StaticHyperlink::StaticHyperlink(HWND hwnd)
 : SubclassedWindow(hwnd)
{
    font = NULL;
    finger = LoadCursor(0, MAKEINTRESOURCE(IDC_HAND));
    if (finger == NULL) {
        finger = LoadCursor(GetModuleHandle(0), MAKEINTRESOURCE(IDC_ARROW));
    }
    color = RGB(0, 0, 255);
}

StaticHyperlink::~StaticHyperlink()
{
    if (font) {
        DeleteObject(font);
    }
}

void StaticHyperlink::SetText(const char text[])
{
    SetWindowText(Hwnd, text);
}

HBRUSH StaticHyperlink::OnCtlColorStatic(HWND parent, HDC dc, HWND control, int type)
{
    HBRUSH hbr = NULL;
    if ((GetWindowLong(Hwnd, GWL_STYLE) & 0xFF) <= SS_RIGHT) {
        if (font == NULL) {
            LOGFONT lf;
            GetObject((HFONT)SendMessage(Hwnd, WM_GETFONT, 0, 0), sizeof(lf), &lf);
            lf.lfUnderline = TRUE;
            font = CreateFontIndirect(&lf);
        }
        SelectObject(dc, font);
        SetTextColor(dc, color);
        SetBkMode(dc, TRANSPARENT);
        hbr = (HBRUSH)GetStockObject(HOLLOW_BRUSH);
    }
    return hbr;
}

void StaticHyperlink::OnLButtonDown(HWND hwnd, BOOL doubleclick, int x, int y, UINT flags)
{
    char buf[256];
    GetWindowText(Hwnd, buf, sizeof(buf));
    HCURSOR cur = GetCursor();
    SetCursor(LoadCursor(NULL, IDC_WAIT));
    HINSTANCE r = ShellExecute(NULL, "open", buf, NULL, NULL, SW_SHOWNORMAL);
    SetCursor(cur);
    if ((UINT)r > 32) {
        color = RGB(128, 0, 128);
        InvalidateRect(Hwnd, NULL, FALSE);
    }
}

UINT StaticHyperlink::OnNcHitTest(HWND hwnd, int x, int y)
{
    return HTCLIENT;
}

BOOL StaticHyperlink::OnSetCursor(HWND hwnd, HWND hwndCursor, UINT codeHitTest, UINT msg)
{
    SetCursor(finger);
    return TRUE;
}

LRESULT StaticHyperlink::WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_CTLCOLORSTATIC:
        return HANDLE_WM_CTLCOLORSTATIC(hwnd, wparam, lparam, OnCtlColorStatic);
    case WM_LBUTTONDOWN:
        return HANDLE_WM_LBUTTONDOWN(hwnd, wparam, lparam, OnLButtonDown);
    case WM_NCHITTEST:
        return HANDLE_WM_NCHITTEST(hwnd, wparam, lparam, OnNcHitTest);
    case WM_SETCURSOR:
        return HANDLE_WM_SETCURSOR(hwnd, wparam, lparam, OnSetCursor);
    default:
        return CallWindowProc(OldWndProc, hwnd, msg, wparam, lparam);
    }
    return 0;
}
