#include <windows.h>
#include <wininet.h>
#include <stdio.h>
#include <time.h>
#pragma hdrstop

#pragma warning(disable: 4786) // debug info too long

#include <string>
#include <map>

#include "resource.h"

#include "regex.h"
#include "registry.h"
#include "statichyperlink.h"
#include "xmldom.h"
#include "version.h"

#define for if(0);else for

using namespace std;

const int WM_TASKBAR = WM_USER + 3333;
const int WM_ICONUPDATE = WM_USER + 3334;

const int uTaskbarCreatedMsg = RegisterWindowMessage("TaskbarCreated");

HINSTANCE hInstance;
HWND hWnd;
HANDLE GetDataEvent;
HICON CurrentIcon = 0;

enum ThreatType {
    UNKNOWN,
    LOW,
    GUARDED,
    ELEVATED,
    HIGH,
    SEVERE
};

const char *ThreatNames[][6] = {
    {
        "Unknown",
        "Low (green)",
        "Guarded (blue)",
        "Elevated (yellow)",
        "High (orange)",
        "Severe (red)"
    },
    {
        "Unknown",
        "Huge (off-white)",
        "Immense (cream)",
        "Enormous (putty)",
        "Gigantic (bone)",
        "Enormous (natural)"
    }
};

struct {
    const char *name;
    int icon;
    HICON icons[6];
} ThreatStyles[] = {
    {
        "Orb",
        IDI_THREAT_ORB
    },
    {
        "USA",
        IDI_THREAT_USA
    },
    {
        "Patriot",
        IDI_THREAT_PAT
    },
    {
        "Flag",
        IDI_THREAT_FLG
    },
    {
        "Shapes",
        IDI_THREAT_SHP
    },
    {
        "Duct tape",
        IDI_THREAT_DCT
    },
};

enum ThreatType Threat, LastThreat;
int ThreatStyle;
time_t LastUpdate;
bool Blinking;
bool BlinkState;
time_t ManualUpdateRequest;
time_t ManualUpdateLastUpdate;
bool SnLmode;

const char *DefaultRegistryKey = "Software\\hewgill.com\\Homeland Security Threat Monitor";

struct Configuration {
    string LatestVersion;
    string OfficialSite;
    string MonitorUrl;
    string MonitorRegex;
    map<string, ThreatType> Match;
};

const char *DefaultParameters =
    "<?xml version=\"1.0\"?>"
    "<threat>"
    "  <latest-version>"THREAT_VERSION"</latest-version>"
    "  <official-site>http://www.dhs.gov/dhspublic/</official-site>"
    "  <monitor>"
    "    <url>http://www.dhs.gov/dhspublic/getAdvisoryCondition</url>"
    "    <regex>CONDITION=\"([A-Z]+)\"</regex>"
    "    <match>"
    "      <item>LOW</item>"
    "      <item>GUARDED</item>"
    "      <item>ELEVATED</item>"
    "      <item>HIGH</item>"
    "      <item>SEVERE</item>"
    "    </match>"
    "  </monitor>"
    "</threat>";

Configuration Parameters;

void LoadSettings()
{
    DWORD x;
    if (GetRegistryDword("ThreatLevel", x) && x >= UNKNOWN && x <= SEVERE) {
        Threat = (ThreatType)x;
        LastThreat = Threat;
    }
    if (GetRegistryDword("IconStyle", x) && x >= 0 && x < sizeof(ThreatStyles)/sizeof(ThreatStyles[0])) {
        ThreatStyle = x;
    }
    if (GetRegistryDword("LastUpdate", x) && LastUpdate < time(0)) {
        LastUpdate = x;
    }
    char s[256];
    if (GetRegistryString("LatestVersion", s, sizeof(s))) {
        Parameters.LatestVersion = s;
    }
    if (GetRegistryString("OfficialSite", s, sizeof(s))) {
        Parameters.OfficialSite = s;
    }
    if (GetRegistryString("MonitorUrl", s, sizeof(s))) {
        Parameters.MonitorUrl = s;
    }
    if (GetRegistryString("MonitorRegex", s, sizeof(s))) {
        Parameters.MonitorRegex = s;
    }
    for (int i = LOW; i <= SEVERE; i++) {
        char buf[20];
        _snprintf(buf, sizeof(buf), "Match%d", i);
        if (GetRegistryString(buf, s, sizeof(s))) {
            Parameters.Match[s] = (ThreatType)i;
        }
    }
}

void SaveSettings()
{
    SetRegistryDword("ThreatLevel", Threat);
    SetRegistryDword("IconStyle", ThreatStyle);
    SetRegistryDword("LastUpdate", LastUpdate);
    SetRegistryString("LatestVersion", Parameters.LatestVersion.c_str());
    SetRegistryString("OfficialSite", Parameters.OfficialSite.c_str());
    SetRegistryString("MonitorUrl", Parameters.MonitorUrl.c_str());
    SetRegistryString("MonitorRegex", Parameters.MonitorRegex.c_str());
    for (map<string, ThreatType>::iterator i = Parameters.Match.begin(); i != Parameters.Match.end(); i++) {
        char buf[20];
        _snprintf(buf, sizeof(buf), "Match%d", i->second);
        SetRegistryString(buf, i->first.c_str());
    }
}

void LoadIcons()
{
    for (int i = 0; i < sizeof(ThreatStyles)/sizeof(ThreatStyles[0]); i++) {
        for (int j = 0; j < 6; j++) {
            ThreatStyles[i].icons[j] = (HICON)LoadImage(hInstance, MAKEINTRESOURCE(ThreatStyles[i].icon+j), IMAGE_ICON, 16, 16, 0);
        }
    }
}

Configuration LoadParameters(const char *parameters)
{
    Configuration c;
    xml::Document params = xml::parseDocumentText(parameters);
    const xml::Element *root = params.getRootElement();
    if (root == NULL) {
        if (parameters == DefaultParameters) {
            MessageBox(NULL, "There was a fatal error loading the configuration parameters. Please visit http://www.hewgill.com/threat/ for a newer version.", "Homeland Security Threat Monitor", MB_OK|MB_ICONSTOP);
            exit(1);
        }
        return LoadParameters(DefaultParameters);
    }
    const xml::Element *e = root->getFirstChild("latest-version");
    if (e == NULL) {
        return LoadParameters(DefaultParameters);
    }
    c.LatestVersion = e->getText();
    e = root->getFirstChild("official-site");
    if (e == NULL) {
        return LoadParameters(DefaultParameters);
    }
    c.OfficialSite = e->getText();
    const xml::Element *monitor = root->getFirstChild("monitor");
    if (monitor == NULL) {
        return LoadParameters(DefaultParameters);
    }
    e = monitor->getFirstChild("url");
    if (e == NULL) {
        return LoadParameters(DefaultParameters);
    }
    c.MonitorUrl = e->getText();
    e = monitor->getFirstChild("regex");
    if (e == NULL) {
        return LoadParameters(DefaultParameters);
    }
    c.MonitorRegex = e->getText();
    const xml::Element *match = monitor->getFirstChild("match");
    if (match == NULL) {
        return LoadParameters(DefaultParameters);
    }
    vector<const xml::Element *> items = match->getElements("item");
    if (items.size() != 5) {
        return LoadParameters(DefaultParameters);
    }
    for (int t = 0; t < 5; t++) {
        c.Match[items[t]->getText()] = (ThreatType)(t+LOW);
    }
    return c;
}

BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_COMMAND:
        switch (LOWORD(wparam)) {
        case IDOK:
        case IDCANCEL:
            EndDialog(hwnd, LOWORD(wparam));
            break;
        }
        break;
    case WM_CTLCOLORSTATIC:
        return SendMessage((HWND)lparam, msg, wparam, lparam);
    case WM_INITDIALOG: {
        new StaticHyperlink(GetDlgItem(hwnd, IDC_WEBSITE));
        StaticHyperlink *official = new StaticHyperlink(GetDlgItem(hwnd, IDC_OFFICIAL));
        official->SetText(Parameters.OfficialSite.c_str());
        SetDlgItemText(hwnd, IDC_THIS_VERSION, THREAT_VERSION);
        SetDlgItemText(hwnd, IDC_LATEST_VERSION, Parameters.LatestVersion.c_str());
        break;
    }
    default:
        return FALSE;
  }
  return TRUE;
}

string GetUrl(HINTERNET session, const char url[], bool nocache = false)
{
    string r;
    HINTERNET request = InternetOpenUrl(session, url, NULL, 0, (nocache ? (INTERNET_FLAG_PRAGMA_NOCACHE|INTERNET_FLAG_NO_CACHE_WRITE) : 0) | INTERNET_FLAG_KEEP_CONNECTION, 0);
    if (request == NULL) {
        return r;
    }
    DWORD code;
    DWORD codelen = sizeof(code);
    if (!HttpQueryInfo(request, HTTP_QUERY_STATUS_CODE|HTTP_QUERY_FLAG_NUMBER, &code, &codelen, 0) || code != 200) {
        InternetCloseHandle(request);
        return r;
    }
    while (true) {
        char buf[4096+1];
        DWORD n;
        if (!InternetReadFile(request, buf, sizeof(buf)-1, &n) || n == 0) {
            break;
        }
        buf[n] = 0;
        r += buf;
    }
    InternetCloseHandle(request);
    return r;
}

DWORD CALLBACK GetThreatDataThread(void *)
{
    bool ConfigLoaded = false;
    while (WaitForSingleObject(GetDataEvent, INFINITE) == WAIT_OBJECT_0) {
        DWORD flags;
        while (!InternetGetConnectedState(&flags, 0)) {
            Sleep(60*1000);
            SendMessage(hWnd, WM_ICONUPDATE, 0, 0);
        }
        HINTERNET session = InternetOpen("Homeland Security Threat Monitor", INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0);
        if (!ConfigLoaded) {
            string params = GetUrl(session, "http://www.hewgill.com/threat/config.xml");
            if (!params.empty()) {
                Parameters = LoadParameters(params.c_str());
            }
            ConfigLoaded = true;
        }
        string page = GetUrl(session, Parameters.MonitorUrl.c_str(), true);
        if (!page.empty()) {
            ThreatType threat = UNKNOWN;
            CRegExp re(Parameters.MonitorRegex.c_str());
            if (re.RegFind(page.c_str()) >= 0) {
                map<string, ThreatType>::iterator mi = Parameters.Match.find(re[1]);
                if (mi != Parameters.Match.end()) {
                    threat = mi->second;
                }
            }
            if (threat != UNKNOWN) {
                Threat = threat;
                SaveSettings();
            }
        }
        InternetCloseHandle(session);
        LastUpdate = time(0);
        ResetEvent(GetDataEvent);
        SendMessage(hWnd, WM_ICONUPDATE, 0, 0);
        SetProcessWorkingSetSize(GetCurrentProcess(), -1, -1);
    }
    return 0;
}

const char *GetThreatName(ThreatType threat)
{
    return ThreatNames[SnLmode][threat];
}

HICON GetThreatIcon(ThreatType threat)
{
    return ThreatStyles[ThreatStyle].icons[SnLmode ? 0 : threat];
}

bool Acknowledge(HWND w)
{
    if (!Blinking) {
        return false;
    }
    Blinking = false;
    SendMessage(w, WM_ICONUPDATE, 0, 0);
    char buf[200];
    _snprintf(buf, sizeof(buf), "The Homeland Security threat level has been changed to \"%s\" by the Attorney General of the United States. Press OK to view the Department of Homeland Security web site.", GetThreatName(Threat));
    if (MessageBox(w, buf, "Homeland Security", MB_OKCANCEL|MB_ICONWARNING) == IDOK) {
        SendMessage(w, WM_TASKBAR, 0, WM_LBUTTONDBLCLK);
    }
    return true;
}

LRESULT CALLBACK WndProc(HWND w, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_COMMAND:
        switch (wparam) {
        case 3000:
            DestroyWindow(w);
            break;
        case 3001:
            return WndProc(w, WM_TASKBAR, 0, WM_LBUTTONDBLCLK);
        case 3002:
            DialogBox(hInstance, MAKEINTRESOURCE(IDD_ABOUT), w, AboutDlgProc);
            break;
        case 3003:
            ManualUpdateRequest = time(0);
            ManualUpdateLastUpdate = LastUpdate;
            SetEvent(GetDataEvent);
            break;
        case 3100:
        case 3101:
        case 3102:
        case 3103:
        case 3104:
        case 3105:
            ThreatStyle = wparam - 3100;
            SaveSettings();
            return WndProc(w, WM_ICONUPDATE, 0, 0);
        case 4000:
            MessageBox(w, "The threat level menu options are provided for reference only. Only the Attorney General of the United States may change the current homeland security threat level.", "Homeland Security", MB_OK|MB_ICONSTOP);
            break;
        }
        break;
    case WM_CREATE:
        NOTIFYICONDATA nid;
        nid.cbSize = sizeof(nid);
        nid.hWnd = w;
        nid.uID = 1000;
        nid.uFlags = NIF_ICON | NIF_MESSAGE;
        nid.uCallbackMessage = WM_TASKBAR;
        nid.hIcon = GetThreatIcon(Threat);
        Shell_NotifyIcon(NIM_ADD, &nid);
        SetTimer(w, 1, 1000, NULL);
        break;
    case WM_DESTROY: {
        NOTIFYICONDATA nid;
        nid.cbSize = sizeof(nid);
        nid.hWnd = w;
        nid.uID = 1000;
        nid.uFlags = 0;
        Shell_NotifyIcon(NIM_DELETE, &nid);
        KillTimer(w, 1);
        PostQuitMessage(0);
        break;
    }
    case WM_TASKBAR:
        if (lparam == WM_LBUTTONDOWN) {
            Acknowledge(w);
        } else if (lparam == WM_LBUTTONDBLCLK) {
            HCURSOR cur = GetCursor();
            SetCursor(LoadCursor(NULL, IDC_WAIT));
            const char *page;
            if (SnLmode) {
                page = "http://snltranscripts.jt.org/01/01oridge.phtml";
            } else {
                page = Parameters.OfficialSite.c_str();
            }
            HINSTANCE r = ShellExecute(NULL, "open", page, NULL, NULL, SW_SHOWNORMAL);
            SetCursor(cur);
        } else if (lparam == WM_RBUTTONDOWN) {
            if (Acknowledge(w)) {
                break;
            }
            SetForegroundWindow(w);
            SetCapture(w);
            HMENU mm = CreatePopupMenu();
            for (int i = 0; i < sizeof(ThreatStyles)/sizeof(ThreatStyles[0]); i++) {
                AppendMenu(mm, MF_STRING, 3100+i, ThreatStyles[i].name);
            }
            CheckMenuRadioItem(mm, 3100, 3100+sizeof(ThreatStyles)/sizeof(ThreatStyles[0])-1, 3100+ThreatStyle, MF_BYCOMMAND);
            HMENU m = CreatePopupMenu();
            AppendMenu(m, MF_SEPARATOR, 0, NULL);
            AppendMenu(m, MF_STRING|MF_POPUP, (UINT)mm, "Style");
            AppendMenu(m, MF_STRING, 3003, "Update now");
            AppendMenu(m, MF_STRING, 3001, "Homeland Security");
            AppendMenu(m, MF_STRING, 3002, "About");
            AppendMenu(m, MF_STRING, 3000, "Close");
            MENUITEMINFO mii;
            mii.cbSize = sizeof(mii);
            mii.fMask = MIIM_ID | MIIM_STATE | MIIM_TYPE;
            mii.fType = MFT_STRING | MFT_RADIOCHECK;
            mii.wID = 4000;
            for (int t = LOW; t <= SEVERE; t++) {
                mii.fState = (Threat == t ? MFS_CHECKED : 0);
                mii.dwTypeData = const_cast<char *>(GetThreatName((ThreatType)t));
                InsertMenuItem(m, 0, TRUE, &mii);
            }
            POINT p;
            GetCursorPos(&p);
            TrackPopupMenu(m, TPM_LEFTALIGN|TPM_RIGHTBUTTON, p.x, p.y, 0, w, NULL);
            ReleaseCapture();
            DestroyMenu(m);
        } else if (lparam == WM_RBUTTONDBLCLK) {
            SnLmode = !SnLmode;
            return WndProc(w, WM_ICONUPDATE, 0, 0);
        }
        break;
    case WM_TIMER:
        if (ManualUpdateRequest > 0 && time(0) - ManualUpdateRequest > 60) {
            ManualUpdateRequest = 0;
            time_t oldlastupdate = LastUpdate;
            if (MessageBox(w, "A requested manual update has not completed within one minute. Do you wish to continue trying?", "Homeland Security", MB_YESNO|MB_ICONQUESTION) == IDYES) {
                if (LastUpdate == oldlastupdate) {
                    ManualUpdateRequest = time(0);
                    SetEvent(GetDataEvent);
                }
            }
        }
        if (time(0) - LastUpdate > 7*3600) {
            SetEvent(GetDataEvent);
        }
        if (Blinking) {
            BlinkState = !BlinkState;
            NOTIFYICONDATA nid;
            nid.cbSize = sizeof(nid);
            nid.hWnd = w;
            nid.uID = 1000;
            nid.uFlags = NIF_ICON;
            nid.hIcon = BlinkState ? GetThreatIcon(Threat) : GetThreatIcon(UNKNOWN);
            Shell_NotifyIcon(NIM_MODIFY, &nid);
        }
        break;
    case WM_ICONUPDATE: {
        if (LastThreat != UNKNOWN && Threat > LastThreat) {
            Blinking = true;
        }
        LastThreat = Threat;
        NOTIFYICONDATA nid;
        nid.cbSize = sizeof(nid);
        nid.hWnd = w;
        nid.uID = 1000;
        nid.uFlags = NIF_ICON | NIF_TIP;
        nid.hIcon = GetThreatIcon(Threat);
        _snprintf(nid.szTip, sizeof(nid.szTip), GetThreatName(Threat));
        if (LastUpdate == 0) {
            strcat(nid.szTip, " (not connected yet)");
        } else {
            int days = (time(0) - LastUpdate) / 86400;
            if (days > 0) {
                _snprintf(&nid.szTip[strlen(nid.szTip)], sizeof(nid.szTip)-strlen(nid.szTip), " (%d day%s ago)", days, days != 1 ? "s" : "");
            }
        }
        Shell_NotifyIcon(NIM_MODIFY, &nid);
        if (ManualUpdateRequest > 0 && LastUpdate > ManualUpdateLastUpdate) {
            ManualUpdateRequest = 0;
            char buf[200];
            _snprintf(buf, sizeof(buf), "Manual update completed successfully. The current threat level is \"%s\".", GetThreatName(Threat));
            MessageBox(w, buf, "Homeland Security", MB_OK|MB_ICONINFORMATION);
        }
        break;
    }
    default:
        if (msg == uTaskbarCreatedMsg) {
            NOTIFYICONDATA nid;
            nid.cbSize = sizeof(nid);
            nid.hWnd = w;
            nid.uID = 1000;
            nid.uFlags = NIF_ICON | NIF_MESSAGE;
            nid.uCallbackMessage = WM_TASKBAR;
            nid.hIcon = GetThreatIcon(Threat);
            Shell_NotifyIcon(NIM_ADD, &nid);
            return WndProc(w, WM_ICONUPDATE, 0, 0);
        }
        return DefWindowProc(w, msg, wparam, lparam);
  }
  return 0;
}

int CALLBACK WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR, int)
{
    ::hInstance = hInstance;
    if (FindWindow("Homeland Security Threat Monitor", "Homeland Security Threat Monitor") != 0) {
        return 0;
    }
    LoadIcons();
    Parameters = LoadParameters(DefaultParameters);
    LoadSettings();
    GetDataEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    DWORD tid;
    CreateThread(NULL, 0, GetThreatDataThread, NULL, 0, &tid);
    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = WndProc;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_THREAT));
    wc.lpszClassName = "Homeland Security Threat Monitor";
    RegisterClass(&wc);
    hWnd = CreateWindow(wc.lpszClassName, "Homeland Security Threat Monitor", 0, 0, 0, 0, 0, NULL, NULL, hInstance, NULL);
    SetEvent(GetDataEvent);
    MSG msg;
    while (GetMessage(&msg, 0, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    CloseHandle(GetDataEvent);
    SaveSettings();
    return 0;
}
